part of 'application.dart';

/// Application root widget
class Application extends StatefulWidget {
  final String name;
  final InitializeApp? initialize;
  final String initialRoute;
  final Set<ActivityRoute> routes;
  final OnUnknownRoute onUnknownRoute;
  final LocaleData locale;
  final AppFactory appFactory;

  const Application._({
    Key? key,
    required this.name,
    required this.initialize,
    required this.initialRoute,
    required this.routes,
    required this.onUnknownRoute,
    required this.locale,
    required this.appFactory,
  }) : super(key: key);

  @override
  ApplicationState createState() => ApplicationState();

  static Future<void> _run({
    required String name,
    required InitializeApp? initialize,
    required Set<ActivityRoute> routes,
    required OnUnknownRoute onUnknownRoute,
    required LocaleData locale,
    required AppFactory appFactory,
  }) async =>
      runApp(Application._(
          key: UniqueKey(),
          name: name,
          initialize: initialize,
          initialRoute: initialize == null ? routes.first.location : (await initialize()).location,
          routes: routes,
          onUnknownRoute: onUnknownRoute,
          locale: locale,
          appFactory: appFactory));

  static ApplicationState of(BuildContext context) =>
      requireNotNull(context.findRootAncestorStateOfType<ApplicationState>(),
          message: () => 'Application not reachable from current context. Have you called runApplication()?');

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(StringProperty('name', name))
      ..add(ObjectFlagProperty<InitializeApp>.has('initialize', initialize))
      ..add(StringProperty('initialRoute', initialRoute))
      ..add(IterableProperty<ActivityRoute>('routes', routes))
      ..add(ObjectFlagProperty<OnUnknownRoute>.has('onUnknownRoute', onUnknownRoute))
      ..add(DiagnosticsProperty<LocaleData>('locale', locale))
      ..add(ObjectFlagProperty<AppFactory>.has('appFactory', appFactory));
  }
}

class ApplicationState extends State<Application> {
  final MutableLiveData<AppLifecycleState> lifecycle =
      MutableLiveData(initialValue: WidgetsBinding.instance!.lifecycleState);

  late String _title;

  String get title => _title;

  set title(String? value) => SystemChrome.setApplicationSwitcherDescription(ApplicationSwitcherDescription(
      label: _title = value ?? widget.name, primaryColor: Theme.of(context).primaryColor.value));

  LocaleData get locale => widget.locale;

  late _AppBindingObserver _binding;

  late AppRouterDelegate _appRouter;

  void restart() {
    _appRouter.clearStorage();
    ViewModelScope.providerOfScope(#application).clear();
    Application._run(
        name: widget.name,
        initialize: widget.initialize,
        routes: widget.routes,
        onUnknownRoute: widget.onUnknownRoute,
        locale: widget.locale,
        appFactory: widget.appFactory);
  }

  @override
  void initState() {
    super.initState();
    _title = widget.name;
    locale._initialize();
    _binding = _AppBindingObserver(this);
    _appRouter = AppRouterDelegate(this, sessionStorage, widget.routes, widget.initialRoute, widget.onUnknownRoute);
  }

  @override
  Widget build(BuildContext context) => ViewModelScope(
      scope: #application,
      child: LiveDataBuilder<Locale?>(
          data: locale.enforcedLocale,
          builder: (context, enforcedLocale) => widget.appFactory(_appRouter.provider, _appRouter.parser, _appRouter,
              (_) => title, enforcedLocale, locale._localizationDelegates, locale.supportedLocales)));

  @override
  void dispose() {
    _binding.dispose();
    super.dispose();
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(DiagnosticsProperty<MutableLiveData<AppLifecycleState>>('lifecycle', lifecycle))
      ..add(DiagnosticsProperty<LocaleData>('locale', locale))
      ..add(StringProperty('title', title));
  }
}
